# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from .user import User
from .escalation_policy import EscalationPolicy
from ..mixins import stringtype


class Team(Entity):
    """PagerDuty team entity."""

    STR_OUTPUT_FIELDS = ('id', 'name',)
    escalationPolicyFactory = EscalationPolicy

    def remove_escalation_policy(self, escalation_policy, **kwargs):
        """Remove an escalation policy from this team."""
        if isinstance(escalation_policy, Entity):
            escalation_policy = escalation_policy['id']

        assert isinstance(escalation_policy, stringtype)

        endpoint = '{0}/{1}/escalation_policies/{2}'.format(
            self.endpoint,
            self['id'],
            escalation_policy,
        )

        return self.request('DELETE', endpoint=endpoint, query_params=kwargs)

    def add_escalation_policy(self, escalation_policy, **kwargs):
        """Add an escalation policy to this team."""
        if isinstance(escalation_policy, Entity):
            escalation_policy = escalation_policy['id']

        assert isinstance(escalation_policy, stringtype)

        endpoint = '{0}/{1}/escalation_policies/{2}'.format(
            self.endpoint,
            self['id'],
            escalation_policy,
        )
        return self.request('PUT', endpoint=endpoint, query_params=kwargs)

    def remove_user(self, user, **kwargs):
        """Remove a user from this team."""
        if isinstance(user, Entity):
            user = user['id']

        assert isinstance(user, stringtype)

        endpoint = '{0}/{1}/escalation_policies/{2}'.format(
            self.endpoint,
            self['id'],
            user,
        )
        return self.request('DELETE', endpoint=endpoint, query_params=kwargs)

    def add_user(self, user, **kwargs):
        """Add a user to this team."""
        if isinstance(user, User):
            user = user['id']

        assert isinstance(user, stringtype)

        endpoint = '{0}/{1}/users/{2}'.format(
            self.endpoint,
            self['id'],
            user,
        )
        result = self.request('PUT', endpoint=endpoint, query_params=kwargs)
        return result

    def update(self, *args, **kwargs):
        """Update this team."""
        raise NotImplemented



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity


class AddOn(Entity):
    """PagerDuty add-on entity."""

    endpoint = 'addons'

    @classmethod
    def install(cls, type_, name, src, *args, **kwargs):
        """Install an add-on to this account."""
        data = kwargs.pop('data', None)
        if data is None:
            data = {
                'addon': {
                    'type': type_,
                    'name': name,
                    'src': src,
                }
            }
        cls.create(data=data, *args, **kwargs)

    def update(self, *args, **kwargs):
        """Update this add-on."""
        raise NotImplemented



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from .integration import Integration
from .vendor import Vendor


class Service(Entity):
    """PagerDuty service entity."""

    STR_OUTPUT_FIELDS = ('id', 'name',)
    integrationFactory = Integration
    vendorFactory = Vendor

    ALLOWED_SERVICE_TYPES = [
        # 'service',
        'service_reference',
    ]

    @classmethod
    def validate(cls, service_info):
        """Validate `service_info` to be acceptable service data."""
        assert isinstance(service_info, dict)
        assert (service_info['type'] in cls.ALLOWED_SERVICE_TYPES)

    def create_integration(self, integration_info, **kwargs):
        """
        Create an integration for this incident.

        See: https://v2.developer.pagerduty.com/v2/page/api-reference#!/
              Services/post_services_id_integrations
        """
        service_info = integration_info.get('service')
        vendor_info = integration_info.get('vendor')

        if service_info is not None:
            self.__class__.validate(service_info)

        if vendor_info is not None:
            self.vendorFactory.validate(vendor_info)

        endpoint = '{0}/{1}/integrations'.format(
            self.endpoint,
            self['id'],
        )
        return self.integrationFactory.create(
            endpoint=endpoint,
            api_key=self.api_key,
            data=integration_info,
            query_params=kwargs
        )

    def integrations(self, **kwargs):
        """Retrieve all this services integrations."""
        ids = [ref['id'] for ref in self['integrations']]
        return [Integration.fetch(id, service=self, query_params=kwargs) for id in ids]

    def get_integration(self, id, **kwargs):
        """Retrieve a single integration by id."""
        return Integration.fetch(id, service=self, query_params=kwargs)

    def update_integration(self, *args, **kwargs):
        """Update this integration on this service."""
        raise NotImplemented

    # sugar-pills
    view_integration = get_integration
    add_integration = create_integration



# Copyright (c) PagerDuty.
# See LICENSE for details.
import re
import json
import unittest
import os.path
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
from operator import itemgetter

import requests_mock

from pypd import EscalationPolicy


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i: i + n]


class EntityTestCase(unittest.TestCase):

    def setUp(self):
        self.base_url = 'https://api.pagerduty.com'
        self.endpoint = 'escalation_policies'
        self.url = '%s/%s' % (self.base_url, self.endpoint,)
        self.api_key = 'FAUX_API_KEY'
        self.limit = 25

        base_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
            'data')

        path = os.path.join(base_path, 'sample_escalation_policies.json')
        with open(path) as f:
            self.data = json.load(f)
            self.query_datas = [
                {
                    'limit': self.limit,
                    'offset': n * self.limit,
                    'escalation_policies': chunk,
                }
                for n, chunk in enumerate(chunks(self.data, self.limit))
            ]

        path = os.path.join(base_path, 'sample_services.json')
        with open(path) as f:
            self.service_data = json.load(f)

    @requests_mock.Mocker()
    def test_output_fields(self, m):
        # setup mocked request uris
        query = {
            'limit': self.query_datas[0]['limit'],
            'offset': self.query_datas[0]['offset'],
        }
        url = self.url + '?%s' % urlencode(query)
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)

        eps = EscalationPolicy.find(api_key=self.api_key, fetch_all=True)
        # ensure all the instances are there, as usual
        self.assertEqual(len(eps), len(self.query_datas[0][self.endpoint]))

        for ep in eps:
            # ensure that there is a name field shown for stringified version
            for field in EscalationPolicy.STR_OUTPUT_FIELDS:
                matches = re.findall(r'%s=\"(.+?)\"' % field, str(ep))
                self.assertEqual(matches[0], ep[field])

            matches = re.findall(r'id=\"(.+?)\"', str(ep))
            self.assertEqual(matches[0], ep['id'])

    @requests_mock.Mocker()
    def test_translate_query_params(self, m):
        name = 'Alexis demo EP'
        query = {
            'query': name,
            'limit': self.query_datas[0]['limit'],
            'offset': self.query_datas[0]['offset'],
        }
        url = self.url + '?%s' % urlencode(query)

        # filter the dataset to only include eps with "Alexis" in it
        filtered = list(filter(
            lambda d: d['name'].count('Alexis'),
            self.query_datas[0]['escalation_policies']
        ))
        data = self.query_datas[0].copy()
        data['escalation_policies'] = filtered
        m.register_uri('GET', url, json=data, complete_qs=True)

        # query for the eps
        eps = EscalationPolicy.find(api_key=self.api_key, name=name,)
        self.assertEqual(len(eps), 1)

    @requests_mock.Mocker()
    def test_services(self, m):
        query = {
            'limit': 1,
            'offset': self.query_datas[0]['offset'],
        }
        url = self.url + '?%s' % urlencode(query)
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)

        # get one ep
        ep = EscalationPolicy.find_one(api_key=self.api_key)
        # intersect the escalation policy service data with the service data
        # found in sample_services.json
        service_ids = set(service['id'] for service in self.service_data)
        ep_service_ids = set(service['id'] for service in ep['services'])
        intersect = service_ids.intersection(ep_service_ids)

        # setup fetch url requests with requests_mock using actual data from
        # sample_services.json
        for o in ep['services']:
            url = self.base_url + '/services/' + o['id']
            data = next(iter(filter(
                lambda s: s['id'] in intersect,
                self.service_data
            )))
            data = {'service': data, }
            m.register_uri('GET', url, json=data, complete_qs=True)

        services = ep.services()
        # ensure the number of services were fetched properly
        self.assertEqual(len(services), len(ep['services']))
        # ensure all the proper services are there (by id)
        diff = set(service['id'] for service in services)\
            .difference(set(ep_service_ids))
        self.assertEqual(len(diff), 0)


if __name__ == "__main__":
    unittest.main()




# Copyright (c) PagerDuty.
# See LICENSE for details.
import logging

logger = logging.getLogger('pypd')
verbosity = 1


def set_logger(new_logger):
    """
    Set the global logger for pypd to use.

    Assumes a logging.Logger interface.
    """
    global logger
    logger = new_logger
    return logger


def set_verbosity(level=1):
    """Set logging verbosity level, 0 is lowest."""
    global verbosity
    verbosity = level
    return verbosity


def log(*args, **kwargs):
    """Log things with the global logger."""
    level = kwargs.pop('level', logging.INFO)
    logger.log(level, *args, **kwargs)


def warn(*args, **kwargs):
    kwargs.pop('level', None)
    logger.log(logging.WARNING, *args, **kwargs)


def error(*args, **kwargs):
    kwargs.pop('level', None)
    logger.log(logging.ERROR, *args, **kwargs)


def debug(*args, **kwargs):
    kwargs.pop('level', None)
    logger.log(logging.DEBUG, *args, **kwargs)



# Copyright (c) PagerDuty.
# See LICENSE for details.
"""
Entity module provides a base class Entity for defining a PagerDuty entity.

Entities should be used as the base for all things that ought to be queryable
via PagerDuty v2 API.
"""

from .entity import Entity
from ..mixins import stringtype


class Event(Entity):
    """A model representing V1 Events in PagerDuty's API."""

    base_url = 'https://events.pagerduty.com/generic/2010-04-15/'\
               'create_event.json'
    EVENT_TYPES = ('trigger', 'acknowledge', 'resolve',)

    @classmethod
    def validate(cls, event_info):
        """Validate that provided event information is valid."""
        assert 'service_key' in event_info
        assert isinstance(event_info['service_key'], stringtype)
        assert 'event_type' in event_info
        assert event_info['event_type'] in cls.EVENT_TYPES
        if event_info['event_type'] != cls.EVENT_TYPES[0]:
            assert 'incident_key' in event_info
            assert isinstance(event_info['incident_key'], stringtype)
        else:
            assert 'description' in event_info

        if 'details' in event_info:
            assert isinstance(event_info['details'], dict)

        if 'contexts' in event_info:
            assert isinstance(event_info['contexts'], (list, tuple,))

    @classmethod
    def create(cls, data=None, api_key=None, endpoint=None, add_headers=None,
               **kwargs):
        """Create an event on your PagerDuty account."""
        cls.validate(data)
        inst = cls(api_key=api_key)
        endpoint = ''
        return inst.request('POST',
                            endpoint=endpoint,
                            data=data,
                            query_params=kwargs,
                            add_headers=add_headers,
                            )


class EventV2(Event):
    """Use the PagerDuty V2 Events API."""

    base_url = 'https://events.pagerduty.com/v2/enqueue'
    SEVERITY_TYPES = ('critical', 'error', 'warning', 'info',)

    @classmethod
    def validate(cls, event_info):
        """Validate that provided event information is valid."""
        assert 'routing_key' in event_info
        assert isinstance(event_info['routing_key'], stringtype)
        assert 'event_action' in event_info
        assert event_info['event_action'] in cls.EVENT_TYPES
        assert 'payload' in event_info
        payload = event_info['payload']
        assert payload['summary']
        assert payload['source']
        assert payload['severity'] in cls.SEVERITY_TYPES



# Copyright (c) PagerDuty.
# See LICENSE for details.
"""
Entity module provides a base class Entity for defining a PagerDuty entity.

Entities should be used as the base for all things that ought to be queryable
via PagerDuty v2 API.
"""
import re
try:
    import ujson as json
except ImportError:
    import json

from ..mixins import ClientMixin, stringtype
from ..log import warn


class NotInitialized(Exception):
    """Raise when an entity is not initialized but accessed as if it were."""


class Entity(ClientMixin):
    """
    Base class for implementing a PagerDuty-something.

    Entities can access their underlying entity data by using dict-like
    accessor methods, ie. `get('property_name', default_value)` and using
    entity['property'] accessors.

    Entities inherit from a ClientMixin which handles doing HTTP requests. They
    also have some special properties to them:

    Entity classes have classmethods for which to initialize queries that ought
    to return a class instance, or a list of class instances they are:
        find:
            finds some instances of this entity, returns a list
        find_one:
            finds one instance of this entity, returns instance
        fetch:
            finds one instance of this entity, returns instance
        create:
            creates an instance of this entity (HTTP POST), returns instance
        delete:
            deletes an instance of this entity (HTTP DELETE), returns None

    Entities have instance methods that help interact with an entity on the api
    they are:
        remove:
            deletes this entity instance, returns None
        get:
            access any property on the entity, like dict-accessor get

    Entities have properties that are also interesting (accessed with dot
    notation):
        id:
            the entities ID value (str)
        json:
            the json representation of the entity (dict)
        _data:
            (do not use directly) contains raw entity data, accessible with
            `get(property, default_value)` OR with entity['property'] syntax

    Entity classes use few special class instances:
        TRANSLATE_QUERY_PARAM:
            A list of strings that ought to be translated to 'query' for the
            query string '?query=stuff', eg.

                TRANSLATE_QUERY_PARAM = ('name',)
                query_with_kwargs(name='joeblow')
                # output = 'uri?query=joeblow'
        MAX_LIMIT_VALUE:
            The normal maximum number of entities returned per page from the
            PagerDuty API
        EXCLUDE_FILTERS:
            A list of strings and methods that will be used to filter out
            entities with the matching criteria. Where strings will look
            to match the keyword argument `excludes='joeblow'` to a value found
            on a key matching any EXLCUDE_FILTERS value OR will apply a method
            with signatures `def exclude_method(cls, item, value)`. Eg.

            class CustomEntity(Entity):
                EXCLUDE_FILTERS = ['name', ...]

                # is the same as

                EXCLUDE_FILTERS =
                    [lambda cls, item, ev: item.get('name').count(ev), ...]
    """

    id = None
    _data = None
    parse = None
    endpoint = None
    # flag so subsequent 'save' operations fail, and require a 'clone'
    _is_deleted = False
    EXCLUDE_FILTERS = ('name',)  # exclude will filter on these properties
    STR_OUTPUT_FIELDS = ('id',)  # fields to output in __str__
    TRANSLATE_QUERY_PARAM = ('name',)  # translates uri?query=stuff
    MAX_LIMIT_VALUE = 100

    def __init__(self, api_key=None, _data=None):
        """Initialize Entity model."""
        if _data is not None:
            self._set(_data)

        # sanitize the endpoint name incase people make mistakes
        self.endpoint = self.__class__.get_endpoint()
        if self.endpoint.endswith('/'):
            warn('Endpoints should not end with a trailing slash, %s',
                 self.__class__)
            self.endpoint = self.endpoint[:-1]

        ClientMixin.__init__(self, api_key)

    @staticmethod
    def sanitize_ep(endpoint, plural=False):
        """
        Sanitize an endpoint to a singular or plural form.

        Used mostly for convenience in the `_parse` method to grab the raw
        data from queried datasets.

        XXX: this is el cheapo (no bastante bien)
        """
        # if we need a plural endpoint (acessing lists)
        if plural:
            if endpoint.endswith('y'):
                endpoint = endpoint[:-1] + 'ies'
            elif not endpoint.endswith('s'):
                endpoint += 's'
        else:
            # otherwise make sure it's singular form
            if endpoint.endswith('ies'):
                endpoint = endpoint[:-3] + 'y'
            elif endpoint.endswith('s'):
                endpoint = endpoint[:-1]

        return endpoint

    @classmethod
    def get_endpoint(cls):
        """
        Accessor method to enable omition of endpoint name.

        In general we want the class name to be translated to endpoint name,
        this way unless otherwise specified will translate class name to
        endpoint name.
        """
        if cls.endpoint is not None:
            return cls.endpoint
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', cls.__name__)
        return cls.sanitize_ep(
            re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower(),
            plural=True
        )

    @classmethod
    def _fetch_all(cls, api_key, endpoint=None, offset=0, limit=25, **kwargs):
        """
        Call `self._fetch_page` for as many pages as exist.

        TODO: should be extended to do async page fetches if API allows it via
        exposing total value.

        Returns a list of `cls` instances.
        """
        output = []
        qp = kwargs.copy()
        limit = max(1, min(100, limit))
        maximum = kwargs.get('maximum')
        qp['limit'] = min(limit, maximum) if maximum is not None else limit
        qp['offset'] = offset
        more, total = None, None

        while True:
            entities, options = cls._fetch_page(
                api_key=api_key, endpoint=endpoint, **qp
            )
            output += entities
            more = options.get('more')
            limit = options.get('limit')
            offset = options.get('offset')
            total = options.get('total')

            if more is None:
                if total is None or offset is None:
                    break
                more = (limit + offset) < total

            if not more or (maximum is not None and len(output) >= maximum):
                break

            qp['limit'] = limit
            qp['offset'] = offset + limit

        return output

    @classmethod
    def _fetch_page(cls, api_key, endpoint=None, page_index=0, offset=None,
                    limit=25, **kwargs):
        """
        Fetch a single page of `limit` number of results.

        Optionally provide `page_index` an integer (0-based) index for the
        page to return. Calculated based on `limit` and `offset`.

        Optionally provide `offset` which will override `page_index` if both
        are passed, will be used to calculate the integer offset of items.

        Optionally provide `limit` integer describing how many items pages
        ought to have.

        Returns a tuple containing a list of `cls` instances and response
        options.
        """
        # if offset is provided have it overwrite the page_index provided
        if offset is not None:
            page_index = int(offset / limit)

        # limit can be maximum MAX_LIMIT_VALUE for most PD queries
        limit = max(1, min(cls.MAX_LIMIT_VALUE, limit))

        # make an tmp instance to do query work
        inst = cls(api_key=api_key)

        kwargs['offset'] = int(page_index * limit)
        maximum = kwargs.pop('maximum', None)

        # if maximum is valid, make the limit <= maximum
        kwargs['limit'] = min(limit, maximum) if maximum is not None else limit
        ep = parse_key = cls.sanitize_ep(cls.get_endpoint(), plural=True)

        # if an override to the endpoint is provided use that instead
        # this is useful for nested value searches ie. for
        # `incident_log_entries` but instead of /log_entries querying with
        # context of /incident/INCIDENTID/log_entries.
        # XXX: could be cleaner
        if endpoint is not None:
            ep = endpoint

        response = inst.request('GET', endpoint=ep, query_params=kwargs)
        # XXX: this is a little gross right now. Seems like the best way
        # to do the parsing out of something and then return everything else
        datas = cls._parse(response, key=parse_key)
        response.pop(parse_key, None)
        entities = [cls(api_key=api_key, _data=d) for d in datas]
        # return a tuple
        return entities, response

    @classmethod
    def fetch(cls, id, api_key=None, endpoint=None, add_headers=None,
              **kwargs):
        """
        Fetch a single entity from the API endpoint.

        Used when you know the exact ID that must be queried.
        """
        if endpoint is None:
            endpoint = cls.get_endpoint()

        inst = cls(api_key=api_key)
        parse_key = cls.sanitize_ep(endpoint).split("/")[-1]
        endpoint = '/'.join((endpoint, id))
        data = cls._parse(inst.request('GET',
                                       endpoint=endpoint,
                                       add_headers=add_headers,
                                       query_params=kwargs),
                          key=parse_key)
        inst._set(data)
        return inst

    # sugar-pills
    get = fetch

    @classmethod
    def _find_exclude_filter(cls, excludes, item):
        """
        For each item returned by a `find()` maybe filter it out.

        Called for each item returned to find a function or string to exclude
        `item` from a filtered list. This method should return truthy values
        where `True`-like values will allow `item` to be included into the
        set, and `False` values will not allow them into the set (filter
        predicate).

        This is a dynamic filtering method such that users of the library
        will be able to do something like:

        Class.find(exclude=('email@address.com', 1,))

        Where the coreesponding EXCLUDE_FILTERS = ('email', 'id',). Similar to
        matching any value on any indexed field, where EXCLUDE_FILTERS are the
        indexes.

        XXX: Even explaining this was difficult. Probably an easier more
        pragmatic way to do this.
        """
        # if exclude is left blank (not a list) then the predicate will just
        # be true
        if excludes is None:
            return False

        # oh my...
        def test_each_exclude(exclude_value):
            # excluded_value is one of excludes = (...,)
            def exclude_equals_value_test(exclude_filter):
                # exclude_filter is one of EXCLUDE_FILTERS = (...,)
                if callable(exclude_filter):
                    return exclude_filter(cls, item, exclude_value,)
                return item.get(exclude_filter) == exclude_value
            return any(map(exclude_equals_value_test, cls.EXCLUDE_FILTERS))
        return any(map(test_each_exclude, excludes))

    @classmethod
    def translate_query_params(cls, **kwargs):
        """
        Translate an arbirtary keyword argument to the expected query.

        TODO: refactor this into something less insane.
        XXX: Clean this up. It's *too* flexible.

        In the v2 API, many endpoints expect a particular query argument to be
        in the form of `query=xxx` where `xxx` would be the name of perhaps
        the name, ID or otherwise. This function ought to take a more aptly
        named parameter specified in `TRANSLATE_QUERY_PARAM`, and substitute it
        into the `query` keyword argument. The purpose is so that some models
        (optionally) have nicer named keyword arguments than `query` for easier
        to read python.

        If a query argument is given then the output should be that value. If a
        substitute value is given as a keyword specified in
        `TRANSLATE_QUERY_PARAM`(and query is not) then the `query` argument
        will be that keyword argument.

        Eg. No query param

            TRANSLATE_QUERY_PARAM = ('name',)
            kwargs = {'name': 'PagerDuty',}
            ...
            output = {'query': 'PagerDuty'}

        or, query param explicitly

            TRANSLATE_QUERY_PARAM = ('name',)
            kwargs = {'name': 'PagerDuty', 'query': 'XXXXPlopperDuty'}
            ...
            output = {'query': 'XXXXPlopperDuty'}

        or, TRANSLATE_QUERY_PARAM is None

            TRANSLATE_QUERY_PARAM = None
            kwargs = {'name': 'PagerDuty', 'query': 'XXXXPlopperDuty'}
            ...
            output = {'output': 'XXXXPlopperDuty', 'name': 'PagerDuty'}

        """
        values = []
        output = kwargs.copy()
        query = kwargs.pop('query', None)

        # remove any of the TRANSLATE_QUERY_PARAMs in output
        for param in (cls.TRANSLATE_QUERY_PARAM or []):
            popped = output.pop(param, None)
            if popped is not None:
                values.append(popped)

        # if query is provided, just use it
        if query is not None:
            output['query'] = query
            return output

        # if query is not provided, use the first parameter we removed from
        # the kwargs
        try:
            output['query'] = next(iter(values))
        except StopIteration:
            pass

        return output

    @classmethod
    def find(cls, api_key=None, fetch_all=True, endpoint=None, maximum=None,
             **kwargs):
        """
        Find some entities from the API endpoint.

        If no api_key is provided, the global api key will be used.
        If fetch_all is True, page through all the data and find every record
        that exists.
        If add_headers is provided (as a dict) use it to add headers to the
        HTTP request, eg.

            {'host': 'some.hidden.host'}

        Capitalizing header keys does not matter.

        Remaining keyword arguments will be passed as `query_params` to the
        instant method `request` (ClientMixin).
        """
        exclude = kwargs.pop('exclude', None)

        # if exclude param was passed a a string, list-ify it
        if isinstance(exclude, stringtype):
            exclude = [exclude, ]

        query_params = cls.translate_query_params(**kwargs)

        # unless otherwise specified use the class variable for the endpoint
        if endpoint is None:
            endpoint = cls.get_endpoint()

        if fetch_all:
            result = cls._fetch_all(api_key=api_key, endpoint=endpoint,
                                    maximum=maximum,
                                    **query_params)
        else:
            result = cls._fetch_page(api_key=api_key, endpoint=endpoint,
                                     maximum=maximum,
                                     **query_params)

        # for each result run it through an exlcusion filter
        collection = [r for r in result
                      if not cls._find_exclude_filter(exclude, r)]
        return collection

    @classmethod
    def find_one(cls, *args, **kwargs):
        """Like `find()` except ensure that only one result is returned."""
        # ensure that maximum is supplied so that a big query is not happening
        # behind the scenes
        if 'maximum' not in kwargs:
            kwargs['maximum'] = 1

        # call find and extract the first iterated value from the result
        iterable = iter(cls.find(*args, **kwargs))
        return next(iterable)

    @classmethod
    def create(cls, data=None, api_key=None, endpoint=None, add_headers=None,
               data_key=None, response_data_key=None, method='POST', **kwargs):
        """
        Create an instance of the Entity model by calling to the API endpoint.

        This ensures that server knows about the creation before returning
        the class instance.

        NOTE: The server must return a response with the schema containing
        the entire entity value. A True or False response is no bueno.
        """
        inst = cls(api_key=api_key)

        if data_key is None:
            data_key = cls.sanitize_ep(cls.get_endpoint())

        if response_data_key is None:
            response_data_key = cls.sanitize_ep(cls.get_endpoint())

        body = {}
        body[data_key] = data

        if endpoint is None:
            endpoint = cls.get_endpoint()

        inst._set(cls._parse(inst.request(method,
                                          endpoint=endpoint,
                                          data=body,
                                          query_params=kwargs,
                                          add_headers=add_headers,
                                          ),
                             key=response_data_key))
        return inst

    # sugar-pills
    post = create

    @classmethod
    def delete(cls, id, api_key=None, **kwargs):
        """Delete an entity from the server by ID."""
        inst = cls(api_key=api_key)
        endpoint = '/'.join((cls.get_endpoint(), id))
        inst.request('DELETE', endpoint=endpoint, query_params=kwargs)
        inst._is_deleted = True
        return True

    @classmethod
    def put(cls, id, api_key=None, **kwargs):
        """Delete an entity from the server by ID."""
        inst = cls(api_key=api_key)
        endpoint = '/'.join((cls.get_endpoint(), id))
        return inst.request('PUT', endpoint=endpoint, query_params=kwargs)

    @classmethod
    def _parse(cls, data, key=None):
        """
        Parse a set of data to extract entity-only data.

        Use classmethod `parse` if available, otherwise use the `endpoint`
        class variable to extract data from a data blob.
        """
        parse = cls.parse if cls.parse is not None else cls.get_endpoint()

        if callable(parse):
            data = parse(data)
        elif isinstance(parse, str):
            data = data[key]
        else:
            raise Exception('"parse" should be a callable or string got, {0}'
                            .format(parse))
        return data

    def _set(self, data):
        self._data = data

    @property
    def id(self):
        """
        Return the entity's ID from `self._data`.

        If not initialized raise a `NotInitialized` exception.
        """
        if self._data is None:
            raise NotInitialized
        return self._data['id']

    @property
    def json(self):
        """
        Return a dict that can be serialized to a JSON-string.

        Does NOT return an encoded string.
        """
        return self._data

    def remove(self):
        """Delete this instance from server record."""
        return self.__class__.delete(self.id)

    def __getitem__(self, attr):
        """Attribute accessor method in dict-like fashion."""
        try:
            return self._data[attr]
        except:
            raise AttributeError("'%s' has no attribute '%s'" %
                                 (type(self), attr,))

    def get(self, attr, default=None):
        """Attribute accessor method in dict-like fashion."""
        try:
            return self[attr]
        except:
            return default

    def __json__(self):
        """Return a valid JSON string dump of the entity data."""
        return json.dumps(self._data)

    def __iter__(self):
        """Return an iterable key object."""
        if not self._data:
            return iter({})
        return iter(self._data.copy())

    def __str__(self):
        """Return a more meaningful class string."""
        id_ = hex(id(self))
        clsname = self.__class__.__name__

        info = {}
        for field in self.__class__.STR_OUTPUT_FIELDS:
            depth = field.split('.')
            original_field = field
            try:
                value = self
                while depth:
                    field = depth.pop(0)
                    value = value[field]
            except:
                pass

            info[original_field] = value

        if not info.get('id') and 'id' in self.__class__.STR_OUTPUT_FIELDS:
            return '<%s uninitialized at %s>' % (clsname, id_,)

        output = '<%s ' % clsname
        for k, v in info.items():
            output += '%s="%s" ' % (k, v)
        output += 'at %s>' % id_
        return output

    def __repr__(self):
        """Return a more meaningful programmer representation string."""
        return self.__str__()



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from ..errors import (InvalidEndpointOperation, InvalidEndpoint,
                      InvalidArguments)


class Note(Entity):
    """PagerDuty note entity."""

    @classmethod
    def fetch(*args, **kwargs):
        """Disable this endpoint, not valid v2."""
        raise InvalidEndpoint('Not a valid location on this endpoint')

    def remove(self, *args, **kwargs):
        """Disable this operation, not valid on this endpoint."""
        raise InvalidEndpointOperation(
            'Not a valid operation on this endpoint.'
        )

    delete = fetch

    @classmethod
    def create(cls, incident=None, endpoint=None, *args, **kwargs):
        """
        Create a note within the scope of an incident.

        Make sure that they should reasonably be able to query with an
        incident or endpoint that knows about an incident.
        """
        if incident is None and endpoint is None:
            raise InvalidArguments(incident, endpoint)
        if endpoint is None:
            iid = incident['id'] if isinstance(incident, Entity) else incident
            endpoint = 'incidents/{0}/notes'.format(iid)

        # otherwise endpoint should contain the incident path too
        return getattr(Entity, 'create').__func__(
            cls,
            endpoint=endpoint,
            *args,
            **kwargs
        )



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity


class User(Entity):
    """PagerDuty user entity."""
    STR_OUTPUT_FIELDS = ('id', 'email',)
    EXCLUDE_FILTERS = TRANSLATE_QUERY_PARAM = ('email', 'name',)

    @property
    def email(self):
        return self['email']

    @classmethod
    def create(cls, data=None, api_key=None, add_headers=None, from_email=None,
               **kwargs):
        if add_headers is None:
            add_headers = {}

        if from_email is None:
            raise Exception('%s.create requires from_email keyword argument'
                            % (cls,))
        if from_email is not None:
            add_headers['From'] = from_email

        return getattr(Entity, 'create').__func__(
            cls,
            data=data,
            api_key=api_key,
            add_headers=add_headers
        )

    def contact_methods(self, **kwargs):
        """Get all contact methods for this user."""
        endpoint = '{0}/{1}/contact_methods'.format(
            self.endpoint,
            self['id'],
        )
        result = self.request('GET', endpoint=endpoint, query_params=kwargs)
        return result['contact_methods']

    def create_contact_method(self, data, **kwargs):
        """Create a contact method for this user."""
        data = {'contact_method': data, }
        endpoint = '{0}/{1}/contact_methods'.format(
            self.endpoint,
            self['id'],
        )
        result = self.request('POST', endpoint=endpoint, data=data,
                              query_params=kwargs)
        self._data['contact_methods'].append(result['contact_method'])
        return result

    def delete_contact_method(self, id, **kwargs):
        """Delete a contact method for this user."""
        endpoint = '{0}/{1}/contact_methods/{2}'.format(
            self.endpoint,
            self['id'],
            id,
        )
        return self.request('DELETE', endpoint=endpoint, query_params=kwargs)

    def get_contact_method(self, id, **kwargs):
        """Get a contact method for this user."""
        endpoint = '{0}/{1}/contact_methods/{2}'.format(
            self.endpoint,
            self['id'],
            id,
        )
        result = self.request('GET', endpoint=endpoint, query_params=kwargs)
        return result['contact_method']

    def notification_rules(self, **kwargs):
        """Get all notification rules for this user."""
        endpoint = '{0}/{1}/notification_rules'.format(
            self.endpoint,
            self['id'],
        )
        result = self.request('GET', endpoint=endpoint, query_params=kwargs)
        return result['notification_rules']

    def get_notification_rule(self, id, **kwargs):
        """Get a notification rule for this user."""
        endpoint = '{0}/{1}/notification_rules/{2}'.format(
            self.endpoint,
            self['id'],
            id,
        )
        result = self.request('GET', endpoint=endpoint, query_params=kwargs)
        return result['notification_rule']

    def create_notification_rule(self, data, **kwargs):
        """Create a notification rule for this user."""
        data = {'notification_rule': data, }
        endpoint = '{0}/{1}/notification_rules'.format(
            self.endpoint,
            self['id'],
        )
        result = self.request('POST', endpoint=endpoint, data=data,
                              query_params=kwargs)
        self._data['notification_rules'].append(result['notification_rule'])
        return result

    def delete_notification_rule(self, id, **kwargs):
        """Get a notification rule for this user."""
        endpoint = '{0}/{1}/notification_rules/{2}'.format(
            self.endpoint,
            self['id'],
            id,
        )
        return self.request('DELETE', endpoint=endpoint, query_params=kwargs)

    def update_contact_method(self, *args, **kwargs):
        """Please implement me."""
        raise NotImplemented

    def update_notification_rule(self, *args, **kwargs):
        """Please implement me."""
        raise NotImplemented

    def update(self, *args, **kwargs):
        """Update this team."""
        raise NotImplemented



# Copyright (c) PagerDuty.
# See LICENSE for details.
try:
    import ujson as json
except ImportError:
    import json


class Error(Exception):
    pass


class InvalidHeaders(Error):
    def __init__(self, headers):
        self.headers = headers
        self.message = "Request received an invalid headers object, %r".format(
            headers,
        )


class BadRequest(Error):
    def __init__(self, code=400, message='', *args, **kwargs):
        self.code = code
        self.message = message
        self.errors = []

        Error.__init__(self, *args, **kwargs)

        try:
            obj = json.loads(message)['error']
            self.code = obj.get('code', code)
            self.message = obj.get('message', message)
            self.errors = obj.get('errors', [])
        except:
            pass

    def __str__(self):
        error = '{0} ({1}): {2}'.format(self.__class__.__name__, self.code,
                                        self.message)
        if self.errors:
            addon = self.errors[0] if len(self.errors) == 1 else self.errors
            error = '{0} - {1}'.format(error, addon)

        return error


class UnknownError(Error):
    def __init__(self, code, url, message=''):
        self.code = code
        self.url = url
        self.message = message

    def __str__(self):
        return '{0} ({1}): {2}'.format(self.__class__.__name__, self.code,
                                       self.message)


class InvalidResponse(Error):
    """A server response was pure nonsense."""

    def __init__(self, data):
        self.data = data

    def __str__(self):
        return 'InvalidResponse: {0}'.format(self.data)


class InvalidArguments(Error):
    """A function was passed invalid arguments."""

    def __init__(self, *args):
        """Initialize the exception."""
        self.code = 400
        self.args = args
        Error.__init__(self)

    def __str__(self):
        """Return a stringified error."""
        error = '{0} ({1}): {2}'.format(self.__class__.__name__, self.code,
                                        self.args)
        return error


class InvalidEndpoint(Error):
    """An endpoint was accessed that is not a valid API endpoint."""


class InvalidEndpointOperation(Error):
    """An invalid operation on this endpoint was accessed."""


class MissingFromEmail(Error):
    """A function was called with missing from email header."""

    def __init__(self, from_email):
        """Initialize the exception with 400 status."""
        self.code = 400
        self.from_email = from_email
        Error.__init__(self)

    def __str__(self):
        """Return a stringified error."""
        return '{0} ({1}): Requires \'from_email\' to be valid, got {2}'\
            .format(self.__class__.__name__, self.code, self.from_email,)



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from .service import Service


class EscalationPolicy(Entity):
    """PagerDuty escalation policy entity."""

    STR_OUTPUT_FIELDS = ('id', 'name',)
    TRANSLATE_QUERY_PARAM = ('name',)

    def services(self):
        """Fetch all instances of services for this EP."""
        ids = [ref['id'] for ref in self['services']]
        return [Service.fetch(id) for id in ids]

    def update(self, *args, **kwargs):
        """Update this escalation policy."""
        raise NotImplemented



# Copyright (c) PagerDuty.
# See LICENSE for details.
import json
import unittest
import os.path

import requests_mock

from pypd import Incident, Alert


def get_object_by_id(data, search_id):
    return list(filter(lambda s: s['id'] == search_id, data))[0]


def get_data(data_type, mock_object):
    return {data_type: mock_object}


class AlertTestCase(unittest.TestCase):

    def setUp(self):
        self.base_url = 'https://api.pagerduty.com'
        self.api_key = 'FAUX_API_KEY'
        self.limit = 25
        base_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
            'data'
        )
        path = os.path.join(base_path, 'sample_incidents.json')
        with open(path) as f:
            self.incidents_data = json.load(f)

        path = os.path.join(base_path, 'sample_alerts.json')
        with open(path) as f:
            self.alerts_data = json.load(f)

        self.first_incident_id = 'INCIDENT1'
        self.first_incident = get_object_by_id(
            self.incidents_data,
            self.first_incident_id
        )
        self.first_incident_data = get_data('incident', self.first_incident)

        self.second_incident_id = 'INCIDENT2'
        self.second_incident = get_object_by_id(
            self.incidents_data,
            self.second_incident_id
        )
        self.second_incident_data = get_data('incident', self.second_incident)

        self.first_alert_id = 'ALERT1'
        self.first_alert = get_object_by_id(
            self.alerts_data,
            self.first_alert_id
        )
        self.first_alert_data = get_data('alert', self.first_alert)

        self.second_alert_id = 'ALERT2'
        self.second_alert = get_object_by_id(
            self.alerts_data, self.second_alert_id)
        self.second_alert_data = get_data('alert', self.second_alert)

    # test helpers
    def build_incident_url(self, incident_id):
        return '{0}/incidents/{1}'.format(
            self.base_url,
            incident_id,
        )

    def build_incident_alerts_url(self, incident_id):
        return '{0}/incidents/{1}/alerts'.format(
            self.base_url,
            incident_id,
        )

    def build_alert_url(self, incident_id, alert_id):
        return '{0}/incidents/{1}/alerts/{2}'.format(
            self.base_url,
            incident_id,
            alert_id,
        )

    def mock_request(self, m, method, url, payload):
        m.register_uri(
            method,
            url,
            json=payload,
            complete_qs=False
        )

    def mock_get_request(self, m, url, payload):
        self.mock_request(m, 'GET', url, payload)

    def mock_put_request(self, m, url, payload):
        self.mock_request(m, 'PUT', url, payload)

    @requests_mock.Mocker()
    def test_fetch_all_alerts_from_incident(self, m):
        incident_url = self.build_incident_url(self.first_incident_id)
        self.mock_get_request(m, incident_url, self.first_incident_data)

        alerts_url = self.build_incident_alerts_url(self.first_incident_id)
        self.mock_get_request(
            m,
            alerts_url,
            {'alerts': [self.first_alert, self.second_alert, ]}
        )

        incident = Incident.fetch(self.first_incident_id, api_key=self.api_key)
        alerts = incident.alerts()
        alert_ids = [i.id for i in alerts]

        for alert_id in [self.first_alert_id, self.second_alert_id]:
            self.assertIn(alert_id, alert_ids)

    @requests_mock.Mocker()
    def test_fetch_one_from_incident(self, m):
        incident_url = self.build_incident_url(self.first_incident_id)
        self.mock_get_request(m, incident_url, self.first_incident_data)

        incident_alerts_url = self.build_incident_alerts_url(
            self.first_incident_id
        )
        self.mock_get_request(
            m,
            incident_alerts_url,
            {'alerts': [self.first_alert, self.second_alert, ]}
        )

        alert_url = self.build_alert_url(
            self.first_incident_id,
            self.first_alert_id
        )
        self.mock_get_request(m, alert_url, self.first_alert_data)

        incident = Incident.fetch(self.first_incident_id, api_key=self.api_key)
        alerts = incident.alerts()
        alert = Alert.fetch(
            self.first_alert_id,
            incident,
            None,
            api_key=self.api_key
        )

        incident_alert_ids = [i.id for i in alerts]
        self.assertIn(alert.id, incident_alert_ids)

    @requests_mock.Mocker()
    def test_resolve(self, m):
        incident_url = self.build_incident_url(self.first_incident_id)
        self.mock_get_request(m, incident_url, self.first_incident_data)

        alert_url = self.build_alert_url(
            self.first_incident_id,
            self.first_alert_id
        )
        self.mock_get_request(m, alert_url, self.first_alert_data)
        self.mock_put_request(m, alert_url, self.first_alert_data)

        incident = Incident.fetch(self.first_incident_id, api_key=self.api_key)
        alert = Alert.fetch(
            self.first_alert_id,
            incident,
            None,
            api_key=self.api_key
        )
        alert.resolve('nizar@pagerduty.com')

        last_request_json = m.last_request.json()
        self.assertEqual('PUT', m.last_request.method)
        self.assertEqual(self.first_alert_id, last_request_json['alert']['id'])
        self.assertEqual('resolved', last_request_json['alert']['status'])

    @requests_mock.Mocker()
    def test_associate(self, m):
        incident_url = self.build_incident_url(self.first_incident_id)
        self.mock_get_request(m, incident_url, self.first_incident_data)

        second_incident_url = self.build_incident_url(self.second_incident_id)
        self.mock_get_request(
            m,
            second_incident_url,
            self.second_incident_data
        )

        alert_url = self.build_alert_url(
            self.first_incident_id,
            self.first_alert_id
        )
        self.mock_get_request(m, alert_url, self.first_alert_data)
        self.mock_put_request(m, alert_url, self.first_alert_data)

        incident = Incident.fetch(self.first_incident_id, api_key=self.api_key)
        new_incident = Incident.fetch(
            self.second_incident_id,
            api_key=self.api_key
        )
        alert = Alert.fetch(
            self.first_alert_id,
            incident,
            None,
            api_key=self.api_key
        )

        alert.associate('nizar@pagerduty.com', new_incident)

        last_request_json = m.last_request.json()
        self.assertEqual('PUT', m.last_request.method)
        self.assertEqual(self.first_alert_id, last_request_json['alert']['id'])
        self.assertEqual(
            self.second_incident_id,
            last_request_json['alert']['incident']['id']
        )

if __name__ == "__main__":
    unittest.main()



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity


class MaintenanceWindow(Entity):
    """PagerDuty maintenance window entity."""

    def update(self, *args, **kwargs):
        """Update a MaintenanceWindow."""
        raise NotImplemented



# Copyright (c) PagerDuty.
# See LICENSE for details.

try:
    import ujson as json
except ImportError:
    import json

from .entity import Entity
from ..errors import InvalidArguments, MissingFromEmail


class Alert(Entity):
    @classmethod
    def fetch(cls, id, incident=None, endpoint=None, *args, **kwargs):
        """Customize fetch because this is a nested resource."""
        if incident is None and endpoint is None:
            raise InvalidArguments(incident, endpoint)

        if endpoint is None:
            iid = incident['id'] if isinstance(incident, Entity) else incident
            endpoint = 'incidents/{0}/alerts'.format(iid)

        return getattr(Entity, 'fetch').__func__(cls, id, endpoint=endpoint,
                                                 *args, **kwargs)

    def resolve(self, from_email):
        """Resolve an alert using a valid email address."""
        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        parent_incident_id = self['incident']['id']
        endpoint_format = 'incidents/{0}/alerts/{1}'
        endpoint = endpoint_format.format(parent_incident_id, self['id'])

        add_headers = {'from': from_email, }
        data = {
            'alert': {
                'id': self['id'],
                'type': 'alert',
                'status': 'resolved',
            }
        }

        result = self.request('PUT',
                              endpoint=endpoint,
                              add_headers=add_headers,
                              data=data,)
        return result

    def associate(self, from_email, new_parent_incident=None):
        """Associate an alert with an incident using a valid email address."""
        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        if new_parent_incident is None:
            raise InvalidArguments(new_parent_incident)

        parent_incident_id = self['incident']['id']
        endpoint_format = 'incidents/{0}/alerts/{1}'
        endpoint = endpoint_format.format(parent_incident_id, self['id'])

        if isinstance(new_parent_incident, Entity):
            new_parent_incident_id = new_parent_incident['id']
        else:
            new_parent_incident_id = new_parent_incident

        add_headers = {'from': from_email, }
        data = {
            'alert': {
                'id': self['id'],
                'type': 'alert',
                'incident': {
                    'type': 'incident',
                    'id': new_parent_incident_id,
                }
            }
        }

        result = self.request('PUT',
                              endpoint=endpoint,
                              add_headers=add_headers,
                              data=data,)
        return result

    def update(self, *args, **kwargs):
        """Update an alert."""
        raise NotImplemented



# Copyright (c) PagerDuty.
# See LICENSE for details.
import datetime

from .entity import Entity
from ..errors import (InvalidArguments, InvalidEndpoint,
                      InvalidEndpointOperation)


class Notification(Entity):
    """A PagerDuty Notification entity."""

    @classmethod
    def find(cls, *args, **kwargs):
        """
        Find notifications.

        Optional kwargs are:
            since:
                datetime instance
            until:
                datetime instance

        If not specified, until will default to now(), and since will default
        to 30 days prior to until.

        As per PD spec, date range must not exceed 1 month.
        """
        seconds = 60 * 60 * 24 * 30  # seconds in 30 days
        until = kwargs.pop('until', None)
        since = kwargs.pop('since', None)

        if until is None:
            until = datetime.datetime.now()

        if since is None:
            since = until - datetime.timedelta(seconds=seconds)

        dt = until - since
        if dt > datetime.timedelta(seconds=seconds):
            raise InvalidArguments(until, since)

        kwargs['since'] = since.isoformat()
        kwargs['until'] = until.isoformat()

        return getattr(Entity, 'find').__func__(cls, *args, **kwargs)

    @classmethod
    def fetch(*args, **kwargs):
        """Disable this endpoint, not valid v2."""
        raise InvalidEndpoint('Not a valid location on this endpoint')

    def remove(self, *args, **kwargs):
        """Disable this operation, not valid on this endpoint."""
        raise InvalidEndpointOperation(
            'Not a valid operation on this endpoint.'
        )

    create = fetch
    delete = fetch



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from ..errors import (InvalidEndpointOperation, InvalidArguments,
                      InvalidEndpoint)


class Integration(Entity):
    """PagerDuty note entity."""

    ALLOWED_INTEGRATION_TYPES = [
        'aws_cloudwatch_inbound_integration',
        'aws_cloudwatch_inbound_integration_reference',
        'cloudkick_inbound_integration',
        'cloudkick_inbound_integration_reference',
        'event_transformer_api_inbound_integration',
        'event_transformer_api_inbound_integration_reference',
        'generic_email_inbound_integration',
        'generic_email_inbound_integration_reference',
        'generic_events_api_inbound_integration',
        'generic_events_api_inbound_integration_reference',
        'keynote_inbound_integration',
        'keynote_inbound_integration_reference',
        'nagios_inbound_integration',
        'nagios_inbound_integration_reference',
        'pingdom_inbound_integration',
        'pingdom_inbound_integration_reference',
        'sql_monitor_inbound_integration',
        'sql_monitor_inbound_integration_reference',
    ]

    @classmethod
    def validate(cls, integration_info):
        """Validate `integration_info` to be valid integration data."""
        assert isinstance(integration_info, dict)
        assert (integration_info['type'] in cls.ALLOWED_INTEGRATION_TYPES)

    @classmethod
    def fetch(cls, id, service=None, endpoint=None, *args, **kwargs):
        """Customize fetch because it lives on a special endpoint."""
        if service is None and endpoint is None:
            raise InvalidArguments(service, endpoint)

        if endpoint is None:
            sid = service['id'] if isinstance(service, Entity) else service
            endpoint = 'services/{0}/integrations'.format(sid)

        return getattr(Entity, 'fetch').__func__(cls, id, endpoint=endpoint,
                                                 *args, **kwargs)

    @classmethod
    def delete(*args, **kwargs):
        """Disable this endpoint, not valid v2."""
        raise InvalidEndpoint('Not a valid location on this endpoint')

    def remove(self, *args, **kwargs):
        """Disable this operation, not valid on this endpoint."""
        raise InvalidEndpointOperation(
            'Not a valid operation on this endpoint.'
        )

    @classmethod
    def create(cls, service=None, endpoint=None, data=None, *args, **kwargs):
        """
        Create an integration within the scope of an service.

        Make sure that they should reasonably be able to query with an
        service or endpoint that knows about an service.
        """
        cls.validate(data)

        if service is None and endpoint is None:
            raise InvalidArguments(service, endpoint)
        if endpoint is None:
            sid = service['id'] if isinstance(service, Entity) else service
            endpoint = 'services/{0}/integrations'.format(sid)

        # otherwise endpoint should contain the service path too
        getattr(Entity, 'create').__func__(cls, endpoint=endpoint, data=data,
                                          *args, **kwargs)



# Copyright (c) PagerDuty.
# See LICENSE for details.
import unittest
try:
    from urllib import urlencode
except:
    from urllib.parse import urlencode

import requests_mock

from pypd.models.entity import Entity


class EntityTestCase(unittest.TestCase):

    def setUp(self):
        self.base_url = 'https://api.pagerduty.com'
        self.endpoint = 'entities'
        self.url = '%s/%s' % (self.base_url, self.endpoint)
        self.api_key = 'FAUX_API_KEY'
        self.limit = 1
        self.total = None
        self.responses_data = [
            {
                'limit': self.limit,
                'offset': 0,
                'total': self.total,
                'more': True,
                'entities': [
                    {
                        'id': 'id1234',
                        'name': 'Entity 1'
                    }
                ]
            },
            {
                'limit': self.limit,
                'offset': 1,
                'total': self.total,
                'more': False,
                'entities': [
                    {
                        'id': 'id5678',
                        'name': 'Entity 2',
                    }
                ]
            }
        ]

        class TestEntity(Entity):
            endpoint = self.endpoint

        self.cls = TestEntity

    def test_sanitize_endpoint(self):
        class PluralEndpoint(Entity):
            endpoint = 'plurals'

        class PluralEndsInYEndpoint(Entity):
            endpoint = 'pluraly'

        class PluralEndsNotWithSEndpoint(Entity):
            endpoint = 'plural'

        class SingularEndpoint(Entity):
            endpoint = 'singular'

        class SingularEndpointEndsWithIes(Entity):
            endpoint = 'singularies'

        class SingularEndpointEndswithS(Entity):
            endpoint = 'singulars'

        plural_cases = [
            (PluralEndpoint, 'plurals'),
            (PluralEndsInYEndpoint, 'pluralies'),
            (PluralEndsNotWithSEndpoint, 'plurals'),
        ]

        singular_cases = [
            (SingularEndpoint, 'singular'),
            (SingularEndpointEndsWithIes, 'singulary'),
            (SingularEndpointEndswithS, 'singular'),
        ]

        for cls, expected_result in plural_cases:
            self.assertEqual(
                cls.sanitize_ep(cls.endpoint, plural=True),
                expected_result
            )

        for cls, expected_result in singular_cases:
            self.assertEqual(
                cls.sanitize_ep(cls.endpoint),
                expected_result
            )

    @requests_mock.Mocker()
    def test_fetch_page(self, m):
        method = 'GET'
        limit = 25
        offset = 0
        more = False
        total = None

        # define endpoint entity data
        entities_data = {
            'limit': limit,
            'offset': offset,
            'total': total,
            'more': more,
            'entities': [
                {
                    'id': 'id1234'
                }
            ]
        }

        # register request mock url with resultant data
        m.register_uri(method, self.url, json=entities_data)
        entities, response = self.cls._fetch_page(
            api_key=self.api_key,
        )

        self.assertTrue(isinstance(entities[0], self.cls))
        self.assertEqual(len(entities), len(entities_data['entities']))
        self.assertEqual(response['total'], None)
        self.assertEqual(response['more'], False)
        self.assertEqual(response['limit'], limit)
        self.assertEqual(response['offset'], offset)

    @requests_mock.Mocker()
    def test_fetch_all(self, m):
        # setup mocked request uris
        for data in self.responses_data:
            query = {
                'limit': data['limit'],
                'offset': data['offset'],
            }
            url = self.url + '?%s' % urlencode(query)
            m.register_uri('GET', url, json=data, complete_qs=True)

        # fetch all entities from endpoint
        entities = self.cls._fetch_all(
            api_key=self.api_key,
            limit=self.limit,
        )

        # ensure result is all instances of self.cls and that data on the
        # instances is correct
        for n, entity in enumerate(entities):
            self.assertTrue(isinstance(entity, self.cls))
            self.assertEqual(
                entity['id'],
                self.responses_data[n]['entities'][0]['id']
            )

        # ensure that the result is actually all of the results
        self.assertEqual(
            len(entities),
            len(self.responses_data[0]['entities']) +
            len(self.responses_data[1]['entities'])
        )

    @requests_mock.Mocker()
    def test_find(self, m):
        # setup mocked request uris
        for data in self.responses_data:
            query = {
                'limit': data['limit'],
                'offset': data['offset'],
            }
            url = self.url + '?%s' % urlencode(query)
            m.register_uri('GET', url, json=data, complete_qs=True)

        # try excluding by id
        class ExcludeById(Entity):
            endpoint = 'entities'
            EXCLUDE_FILTERS = ('id',)

        entities = ExcludeById.find(
            api_key=self.api_key,
            limit=self.limit,
            exclude=('id1234',),
        )

        # expect that only 1 (of 2) entities will show up
        self.assertEqual(len(entities), 1)
        # ensure that the 1 result is not the excluded one
        self.assertNotEqual(entities[0]['id'], 'id1234')

        # try excluding by just name
        class ExcludeByName(Entity):
            endpoint = 'entities'

        entities = ExcludeByName.find(
            api_key=self.api_key,
            limit=self.limit,
            exclude=('Entity 2',),
        )

        # expect that only 1 (of 2) entities will show up
        self.assertEqual(len(entities), 1)
        # ensure that the 1 result is not the excluded one
        self.assertNotEqual(entities[0]['name'], 'Entity 2')

        # try exlcuding by id, but only with only 'name' exclude filter
        entities = ExcludeByName.find(
            api_key=self.api_key,
            limit=self.limit,
            exclude=('id1234',),
        )

        # expect that only 1 (of 2) entities will show up
        self.assertEqual(len(entities), 2)

        # try excluding by custom function
        class ExcludeByCallable(Entity):
            endpoint = 'entities'
            EXCLUDE_FILTERS = [
                lambda cls, item, ev: item.get('name').count(ev),
            ]

        entities = ExcludeByCallable.find(
            api_key=self.api_key,
            limit=self.limit,
            exclude=('Entity 1',),
        )

        # expect that only 1 (of 2) entities will show up
        self.assertEqual(len(entities), 1)
        # expect that the excluded one is correctly Entity 1
        self.assertNotEqual(entities[0]['name'], 'Entity 1')

    def test_translate_query_params_with_name(self):
        class TranslateNameQueryParam(Entity):
            TRANSLATE_QUERY_PARAM = ('name',)

        kwargs = {
            'name': 'PagerDuty',
        }

        # ensure fallback query works using TRANSLATE_QUERY_PARAM
        qp = TranslateNameQueryParam.translate_query_params(**kwargs)
        self.assertEqual(qp['query'], kwargs['name'])
        self.assertNotIn('name', qp)


    def test_translate_query_params_with_name_and_query(self):
        class TranslateNameQueryParam(Entity):
            TRANSLATE_QUERY_PARAM = ('name',)

        kwargs = {
            'name': 'PagerDuty',
            'query': 'PlopperDuty',
        }

        qp = TranslateNameQueryParam.translate_query_params(**kwargs)
        self.assertEqual(qp['query'], kwargs['query'])
        self.assertNotIn('name', qp)


    def test_translate_query_params_with_name_and_query_no_translation(self):
        class TranslateNameQueryParam(Entity):
            TRANSLATE_QUERY_PARAM = None

        kwargs = {
            'name': 'PagerDuty',
            'query': 'PlopperDuty',
        }

        qp = TranslateNameQueryParam.translate_query_params(**kwargs)
        self.assertEqual(qp['query'], kwargs['query'])
        self.assertEqual(qp['name'], kwargs['name'])

    @requests_mock.Mocker()
    def test_parse(self, m):
        # setup mocked request uris
        for data in self.responses_data:
            query = {
                'limit': 1,
                'offset': data['offset'],
            }
            url = self.url + '?%s' % urlencode(query)
            m.register_uri('GET', url, json=data, complete_qs=True)

        # define a class to test parsing using parse as a string
        class TestParseString(Entity):
            endpoint = 'entities'
            parse = 'entities'

        # fetch the entities
        entities = TestParseString.find(api_key=self.api_key, maximum=1,)
        # ensure that there is only 1 result (maximum)
        self.assertEqual(len(entities), 1)

        # for each entity ensure instance of TestParseString, and data is valid
        for n, entity in enumerate(entities):
            self.assertEqual(
                entities[n]['id'],
                self.responses_data[n]['entities'][0]['id']
            )
            self.assertTrue(isinstance(entities[n], TestParseString))

        # class to test parse implemented as a classmethod
        class TestParseFunction(Entity):
            endpoint = 'entities'

            @classmethod
            def parse(cls, data):
                return data['entities']

        # fetch entities
        entities = TestParseFunction.find(api_key=self.api_key, limit=1,
                                          maximum=1)
        # ensure only one was returned
        self.assertEqual(len(entities), 1)

        # for each entity ensure instance of TestParseFunction and that
        # returned data is valid
        for n, entity in enumerate(entities):
            self.assertEqual(
                entities[n]['id'],
                self.responses_data[n]['entities'][0]['id']
            )
            self.assertTrue(isinstance(entities[n], TestParseFunction))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) PagerDuty.
# See LICENSE for details.
import json
import unittest
import os.path

import requests_mock

from pypd import Integration, Service
from pypd.errors import (InvalidEndpointOperation, InvalidEndpoint,
                         InvalidArguments)


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i: i + n]


class IntegrationTestCase(unittest.TestCase):

    def setUp(self):
        self.base_url = 'https://api.pagerduty.com'
        self.api_key = 'FAUX_API_KEY'
        self.limit = 25
        base_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
            'data')
        path = os.path.join(base_path, 'sample_services.json')
        with open(path) as f:
            self.service_data = json.load(f)

        path = os.path.join(base_path, 'sample_integrations.json')
        with open(path) as f:
            self.integration_data = json.load(f)

        # use a particular service because of not wanting to enumerate multiple
        # integration endpoints in json
        self.service_id = 'P91VIE2'
        self.service = list(filter(
            lambda s: s['id'] == self.service_id,
            self.service_data,
        ))[0]

        self.service_data = {
            'service': self.service,
        }

        self.integration = list(filter(
            lambda i: i['service']['id'] == self.service_id,
            self.integration_data,
        ))[0]

        self.integration_id = self.integration['id']

    @requests_mock.Mocker()
    def test_fetch_all_from_service(self, m):
        # setup mocked request uris
        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )

        for integration in self.integration_data:
            integration_url = '{0}/integrations/{1}'.format(
                service_url,
                integration['id']
            )

            m.register_uri(
                'GET',
                integration_url,
                json={'integration': integration},
                complete_qs=False
            )

        m.register_uri(
            'GET',
            service_url,
            json=self.service_data,
            complete_qs=False
        )

        service = Service.fetch(self.service_id, api_key=self.api_key)
        integrations = service.integrations()
        integration_ids = [i.id for i in integrations]
        service_data_integration_ids = [
            i['id'] for i in service['integrations']
        ]

        for integration_id in integration_ids:
            self.assertIn(integration_id, service_data_integration_ids)

        for integration in integrations:
            self.assertEqual(integration['service']['id'], self.service_id)

    @requests_mock.Mocker()
    def test_fetch_one_from_service(self, m):
        # setup mocked request uris
        integration_id = 'P1ONR10'
        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )

        integration_data = self.integration_data[0]

        integration_url = '{0}/integrations/{1}'.format(
            service_url,
            integration_data['id']
        )

        m.register_uri(
            'GET',
            integration_url,
            json={'integration': integration_data},
            complete_qs=False
        )

        m.register_uri(
            'GET',
            service_url,
            json=self.service_data,
            complete_qs=False
        )

        service = Service.fetch(self.service_id, api_key=self.api_key)
        integration = service.get_integration(integration_id)
        self.assertEqual(integration['service']['id'], service.id)
        self.assertEqual(
            integration['type'],
            'event_transformer_api_inbound_integration_reference'
        )
        Integration.validate(integration._data)
        self.assertDictEqual(integration._data, integration_data)

    @requests_mock.Mocker()
    def test_fetch_with_service_id(self, m):
        # setup mocked request uris
        integration_id = 'P1ONR10'
        integration_data = self.integration_data[0]

        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )

        integration_url = '{0}/integrations/{1}'.format(
            service_url,
            integration_data['id']
        )

        m.register_uri(
            'GET',
            integration_url,
            json={'integration': integration_data},
            complete_qs=False
        )

        service_id = integration_data['service']['id']
        integration = Integration.fetch(integration_id, service=service_id)
        self.assertDictEqual(integration._data, integration_data)

    @requests_mock.Mocker()
    def test_fetch_with_service(self, m):
        # setup mocked request uris
        integration_id = 'P1ONR10'
        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )

        integration_data = self.integration_data[0]

        integration_url = '{0}/integrations/{1}'.format(
            service_url,
            integration_data['id']
        )

        m.register_uri(
            'GET',
            integration_url,
            json={'integration': integration_data},
            complete_qs=False
        )

        m.register_uri(
            'GET',
            service_url,
            json=self.service_data,
            complete_qs=False
        )

        service = Service.fetch(self.service_id, api_key=self.api_key)
        integration = Integration.fetch(integration_id, service=service)
        self.assertDictEqual(integration._data, integration_data)

    @requests_mock.Mocker()
    def test_delete_and_remove(self, m):
        integration_id = 'P1ONR10'
        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )
        integration_data = self.integration_data[0]

        integration_url = '{0}/integrations/{1}'.format(
            service_url,
            integration_data['id']
        )

        m.register_uri(
            'GET',
            integration_url,
            json={'integration': integration_data},
            complete_qs=False
        )

        integration = Integration.fetch(integration_id,
                                        service=self.service_id)

        with self.assertRaises(InvalidEndpoint):
            Integration.delete(integration_id)

        with self.assertRaises(InvalidEndpointOperation):
            integration.remove()

    @requests_mock.Mocker()
    def test_validate_valid(self, m):
        integration_data = self.integration_data[0]
        Integration.validate(integration_data)

        for allowed_type in Integration.ALLOWED_INTEGRATION_TYPES:
            data = integration_data.copy()
            data['type'] = allowed_type
            Integration.validate(data)

    @requests_mock.Mocker()
    def test_validate_invalid(self, m):
        integration_data = self.integration_data[0].copy()
        integration_data['type'] = 'invalidtype'

        with self.assertRaises(AssertionError):
            Integration.validate(integration_data)

        with self.assertRaises(AssertionError):
            Integration.validate(None)

    @requests_mock.Mocker()
    def test_create_invalid(self, m):
        integration_data = self.integration_data[0]

        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )

        integration_url = '{0}/integrations/{1}'.format(
            service_url,
            integration_data['id']
        )

        m.register_uri(
            'POST',
            integration_url,
            json={'integration': integration_data},
            complete_qs=False
        )

        with self.assertRaises(InvalidArguments):
            Integration.create(service=None, data=integration_data)

    @requests_mock.Mocker()
    def test_create_valid(self, m):
        integration_data = self.integration_data[0].copy()
        integration_data.pop('created_at', None)
        integration_data.pop('id', None)

        service_url = '{0}/services/{1}'.format(
            self.base_url,
            self.service_id,
        )
        create_url = '{0}/integrations'.format(service_url)

        m.register_uri(
            'POST',
            create_url,
            json={'integration': integration_data},
            complete_qs=False
        )

        Integration.create(service=self.service_id, data=integration_data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from ..errors import InvalidEndpointOperation, InvalidEndpoint


class OnCall(Entity):
    endpoint = 'oncalls'  # silly

    STR_OUTPUT_FIELDS = ('escalation_level', 'user.id')

    @classmethod
    def fetch(*args, **kwargs):
        """Disable this endpoint, not valid v2."""
        raise InvalidEndpoint('Not a valid location on this endpoint')

    def remove(self, *args, **kwargs):
        """Disable this operation, not valid on this endpoint."""
        raise InvalidEndpointOperation(
            'Not a valid operation on this endpoint.'
        )

    create = fetch
    delete = fetch



# Copyright (c) PagerDuty.
# See LICENSE for details.
import re
import json
import unittest
import os.path
try:
    from urllib import urlencode
except ImportError:
    from urllib.parse import urlencode
from operator import itemgetter

import requests_mock

from pypd import Incident
from pypd.errors import MissingFromEmail


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i: i + n]


class IncidentTestCase(unittest.TestCase):
    """Tests for Incident model specific cases."""

    def setUp(self):
        """Setup the test case."""
        self.base_url = 'https://api.pagerduty.com'
        self.endpoint = 'incidents'
        self.url = '{0}/{1}'.format(self.base_url, self.endpoint)
        self.api_key = 'FAUX_API_KEY'
        self.limit = 25

        self.base_path = base_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
            'data')

        path = os.path.join(base_path, 'sample_incidents.json')
        with open(path) as f:
            self.data = json.load(f)
            self.query_datas = [
                {
                    'limit': self.limit,
                    'offset': n * self.limit,
                    'incidents': chunk,
                }
                for n, chunk in enumerate(chunks(self.data, self.limit))
            ]

        path = os.path.join(base_path, 'sample_services.json')
        with open(path) as f:
            self.service_data = json.load(f)

    @requests_mock.Mocker()
    def test_resolve_invalid_from_email(self, m):
        """Coverge for using an invalid (cheaply validated) from email."""
        query = {
            'limit': 1,
            'offset': 0,
        }
        url = self.url + '?{}'.format(urlencode(query))
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)
        incident = Incident.find_one(api_key=self.api_key)
        resolution = 'Incident was resolved manually.'

        with self.assertRaises(MissingFromEmail):
            incident.resolve(None)
        with self.assertRaises(MissingFromEmail):
            incident.resolve(None, resolution=resolution)
        with self.assertRaises(MissingFromEmail):
            incident.resolve(1, resolution=resolution)
        with self.assertRaises(MissingFromEmail):
            incident.resolve(incident, resolution=resolution)

    @requests_mock.Mocker()
    def test_resolve_valid(self, m):
        """Coverage for using a valid (cheaply validated) email."""
        query = {
            'limit': 1,
            'offset': 0,
        }
        url = self.url + '?{}'.format(urlencode(query))
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)
        incident = Incident.find_one(api_key=self.api_key)
        path = os.path.join(
            self.base_path,
            'sample_incident_resolve_response.json'
        )
        with open(path) as f:
            resolve_response = json.load(f)
            resolution = resolve_response['incidents'][0]['resolve_reason']

        url = '{0}/{1}'.format(self.url, incident['id'])
        m.register_uri('PUT', url, json=resolve_response, complete_qs=True)
        response = incident.resolve('jdc@pagerduty.com', resolution=resolution)
        self.assertEqual(incident['id'], response['incidents'][0]['id'])

    @requests_mock.Mocker()
    def test_snooze_invalid_from_email(self, m):
        """Coverage for invalid from email for snooze."""
        query = {
            'limit': 1,
            'offset': 0,
        }
        url = self.url + '?{}'.format(urlencode(query))
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)
        incident = Incident.find_one(api_key=self.api_key)
        duration = 3600

        with self.assertRaises(MissingFromEmail):
            incident.snooze(None, duration=duration)
        with self.assertRaises(MissingFromEmail):
            incident.snooze(1, duration=duration)
        with self.assertRaises(MissingFromEmail):
            incident.snooze(incident, duration=duration)

    @requests_mock.Mocker()
    def test_snooze_valid(self, m):
        """Coverage for valid snooze."""
        query = {
            'limit': 1,
            'offset': 0,
        }
        url = self.url + '?{}'.format(urlencode(query))
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)
        incident = Incident.find_one(api_key=self.api_key)
        path = os.path.join(
            self.base_path,
            'sample_incident_snooze_response.json'
        )
        duration = 3600
        with open(path) as f:
            snooze_response = json.load(f)

        url = '{0}/{1}/snooze'.format(self.url, incident['id'])
        m.register_uri('POST', url, json=snooze_response, complete_qs=True)
        snoozed_incident = incident.snooze(
            'jdc@pagerduty.com',
            duration=duration
        )
        self.assertEqual(incident['id'], snoozed_incident['id'])

    @requests_mock.Mocker()
    def test_note_invalid_from_email(self, m):
        """Coverage for invalid from email for notes."""
        query = {
            'limit': 1,
            'offset': 0,
        }
        url = self.url + '?{}'.format(urlencode(query))
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)
        incident = Incident.find_one(api_key=self.api_key)
        content = 'Some notes, derp.'

        with self.assertRaises(MissingFromEmail):
            incident.create_note(None, content=content)
        with self.assertRaises(MissingFromEmail):
            incident.create_note(1, content=content)
        with self.assertRaises(MissingFromEmail):
            incident.create_note(incident, content=content)

    @requests_mock.Mocker()
    def test_create_note_valid(self, m):
        """Coverage for valid snooze."""
        query = {
            'limit': 1,
            'offset': 0,
        }
        url = self.url + '?{}'.format(urlencode(query))
        m.register_uri('GET', url, json=self.query_datas[0], complete_qs=True)
        incident = Incident.find_one(api_key=self.api_key)
        path = os.path.join(
            self.base_path,
            'sample_incident_create_note_response.json'
        )
        with open(path) as f:
            create_response = json.load(f)
            content = create_response['note']['content']

        url = '{0}/{1}/notes'.format(self.url, incident['id'])
        m.register_uri('POST', url, json=create_response, complete_qs=True)
        note = incident.create_note(
            'jdc@pagerduty.com',
            content=content
        )
        self.assertNotEqual(incident['id'], note['id'])
        self.assertEqual(content, note['content'])



# Copyright (c) PagerDuty.
# See LICENSE for details.

if __name__ == '__main__':
    import os.path
    import unittest
    import argparse

    dirname = os.path.dirname(__file__)

    parser = argparse.ArgumentParser(description='Run tests for pypd')
    parser.add_argument('--verbose', '-v', action='count', dest='verbosity',
                        default=1)

    args = parser.parse_args()
    suite = unittest.TestLoader().discover(dirname, pattern='*.py')
    unittest.TextTestRunner(verbosity=args.verbosity).run(suite)



# Copyright (c) PagerDuty.
# See LICENSE for details.
import unittest

import requests_mock
from requests import HTTPError

from pypd.mixins import ClientMixin
from pypd.errors import (BadRequest, UnknownError, InvalidResponse,
                         InvalidHeaders)


# ignore methods that aren't tests
requests_mock.Mocker.TEST_PREFIX = 'test_'


@requests_mock.Mocker()
class ClientMixinTestCase(unittest.TestCase):

    def setUp(self):
        self.base_url = 'https://api.pagerduty.com'
        self.endpoint = 'mock'
        self.url = '%s/%s' % (self.base_url, self.endpoint,)

        class Requester(ClientMixin):
            endpoint = self.endpoint

        self.requester = Requester(base_url=self.base_url)

    def test_request_2xx(self, m):
        method = 'GET'
        body = {'status': 'OK'}
        m.register_uri(method, self.url, json=body)
        result = self.requester.request(method, self.endpoint)
        self.assertEqual(body, result)

    def test_request_4xx(self, m):
        method = 'GET'
        m.register_uri(method, self.url, status_code=401)
        self.assertRaises(
            BadRequest,
            self.requester.request, method, self.endpoint)

    def test_request_404(self, m):
        method = 'GET'
        m.register_uri(method, self.url, status_code=404)

        self.assertRaises(
            HTTPError,
            self.requester.request, method, self.endpoint
        )

    def test_request_5xx(self, m):
        method = 'GET'
        m.register_uri(method, self.url, status_code=500)

        self.assertRaises(
            UnknownError,
            self.requester.request, method, self.endpoint
        )

    def test_request_invalid_json(self, m):
        method = 'GET'
        m.register_uri(method, self.url, text='{"1"}')

        self.assertRaises(
            InvalidResponse,
            self.requester.request, method, self.endpoint
        )

    def test_valid_headers(self, m):
        method = 'GET'
        m.register_uri(method, self.url)
        self.requester.request(method, self.endpoint)

    def test_invalid_headers(self, m):
        method = 'GET'
        m.register_uri(method, self.url)

        self.assertRaises(
            InvalidHeaders,
            self.requester.request, method, self.endpoint,
            headers=''
        )

    def test_statuses_array(self, m):
        method = 'GET'
        body = {'status': 'OK'}
        url = '%s?statuses[]=triggered&statuses[]=acknowledged' % self.url
        m.register_uri(method, url, json=body)
        result = self.requester.request(method, self.endpoint,
                                        query_params={'statuses': ['triggered',
                                                                   'acknowledged']})
        self.assertEqual(body, result)

if __name__ == '__main__':
    unittest.main()



# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from ..errors import InvalidEndpointOperation, InvalidEndpoint


class LogEntry(Entity):
    """PagerDuty log entry entity."""
    STR_OUTPUT_FIELDS = ('id', 'type',)

    @classmethod
    def create(*args, **kwargs):
        """Disable this endpoint, not valid v2."""
        raise InvalidEndpoint('Not a valid location on this endpoint')

    def remove(self, *args, **kwargs):
        """Disable this operation, not valid on this endpoint."""
        raise InvalidEndpointOperation(
            'Not a valid operation on this endpoint.'
        )

    delete = create



"""Helpful mixins for PagerDuty entity classes."""
import datetime
import logging
try:
    import ujson as json
except ImportError:
    import json
from numbers import Number
import sys

import requests

from .log import log
from .errors import (BadRequest, UnknownError, InvalidResponse, InvalidHeaders)


CONTENT_TYPE = 'application/vnd.pagerduty+json;version=2'
AUTH_TEMPLATE = 'Token token={0}'
BASIC_AUTH_TEMPLATE = 'Basic {0}'


if sys.version_info[0] < 3:
    stringtype = basestring
else:
    stringtype = str


class ClientMixin(object):
    api_key = None
    base_url = None

    def __init__(self, api_key=None, base_url=None):
        # if no api key is provided try to get one from the packages api_key
        if api_key:
            self.api_key = api_key

        if self.api_key is None:
            from pypd import api_key
            self.api_key = api_key

        if base_url:
            self.base_url = base_url

        if self.base_url is None:
            from pypd import base_url
            self.base_url = base_url

    def _handle_response(self, response):
        if response.status_code == 404:
            response.raise_for_status()
        elif response.status_code // 100 == 4:
            raise BadRequest(response.status_code, response.text)
        elif response.status_code // 100 != 2:
            raise UnknownError(response.status_code, response.text)

        if not response.text:
            return None

        try:
            response = json.loads(response.text)
        except:
            raise InvalidResponse(response.text)

        return response

    def _do_request(self, method, *args, **kwargs):
        """
        Modularized because API was broken.

        Need to be able to inject Mocked response objects here.
        """
        log('Doing HTTP [{3}] request: {0} - headers: {1} - payload: {2}'.format(
            args[0], kwargs.get('headers'), kwargs.get('data'), method,),
            level=logging.DEBUG,)
        requests_method = getattr(requests, method)
        return self._handle_response(requests_method(*args, **kwargs))

    def request(self, method='GET', endpoint='', query_params=None,
                data=None, add_headers=None, headers=None,):
        auth = 'Token token={0}'.format(self.api_key)
        if query_params is None:
            query_params = {}

        if headers is None:
            headers = {
                'Accept': CONTENT_TYPE,
                'Authorization': auth,
                'Content-Type': 'application/json',
            }
        elif not isinstance(headers, dict):
            raise InvalidHeaders(headers)

        if add_headers is not None:
            headers.update(**add_headers)

        for k, v in query_params.copy().items():
            if isinstance(v, stringtype):
                continue
            elif isinstance(v, Number):
                continue
            elif isinstance(v, datetime.datetime):
                query_params[k] = v.isoformat()
            elif isinstance(v, ClientMixin):
                query_params[k] = v['id']
            try:
                iter(v)
            except:
                continue
            key = '%s[]' % k
            query_params.pop(k)
            values = [v_['id'] if isinstance(v_, ClientMixin) else v_
                      for v_ in v]
            query_params[key] = values

        kwargs = {
            'headers': headers,
            'params': query_params,
        }

        if data is not None:
            kwargs['data'] = json.dumps(data)

        return self._do_request(
            method.lower(),
            '/'.join((self.base_url, endpoint)),
            **kwargs
        )



#!/usr/bin/env python
from setuptools import setup, find_packages

with open('pypd/version.py') as version_file:
    exec(compile(version_file.read(), version_file.name, 'exec'))

options = {
    'name': 'pypd',
    'version': __version__,
    'packages': find_packages(),
    'scripts': [],
    'description': 'A python client for PagerDuty API',
    'author': 'JD Cumpson',
    'author_email': 'jdc@pagerduty.com',
    'maintainer': 'JD Cumpson',
    'maintainer_email': 'jdc@pagerduty.com',
    'license': 'MIT',
    'url': 'https://github.com/PagerDuty/pagerduty-api-python-client',
    'download_url': 'https://github.com/PagerDuty/pagerduty-api-python-client/archive/master.tar.gz',
    'classifiers': [
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    'install_requires': ['requests'],
    'tests_require': [],
    'extras_require': {
        'ujson': 'ujson',
    },
    'cmdclass': {}
}

setup(**options)




# Copyright (c) PagerDuty.
# See LICENSE for details.
import logging

from .version import __version__
from .models.ability import can, abilities
from .models.add_ons import AddOn
from .models.escalation_policy import EscalationPolicy
from .models.event import Event, EventV2
from .models.alert import Alert
from .models.incident import Incident
from .models.integration import Integration
from .models.log_entry import LogEntry
from .models.maintenance_window import MaintenanceWindow
from .models.note import Note
from .models.notification import Notification
from .models.on_call import OnCall
from .models.schedule import Schedule
from .models.service import Service
from .models.team import Team
from .models.user import User
from .models.vendor import Vendor

api_key = None
base_url = 'https://api.pagerduty.com'


def set_api_key_from_file(path, set_global=True):
    """Set the global api_key from a file path."""
    with open(path, 'r+b') as f:
        global api_key
        api_key = f.read().strip()
    return api_key



"""
PagerDuty abilities are strings that represent an account ability.

However, abilities don't *really* follow the style of other endpoints so
instead of making a typical endpoint, we'll just expose methods. That can be
called.
"""
# Copyright (c) PagerDuty.
# See LICENSE for details.
from ..mixins import ClientMixin


def abilities(api_key=None, add_headers=None):
    """Fetch a list of permission-like strings for this account."""
    client = ClientMixin(api_key=api_key)
    result = client.request('GET', endpoint='abilities',
                            add_headers=add_headers,)
    return result['abilities']


def can(ability, add_headers=None):
    """Test whether an ability is allowed."""
    client = ClientMixin(api_key=None)
    try:
        client.request('GET', endpoint='abilities/%s' % ability,
                       add_headers=add_headers)
        return True
    except Exception:
        pass
    return False




# Copyright (c) PagerDuty.
# See LICENSE for details.
from .entity import Entity
from ..errors import InvalidEndpointOperation, InvalidEndpoint


class Vendor(Entity):
    """PagerDuty vedor entity."""

    ALLOWED_VENDOR_TYPES = [
        # 'vendor',
        'vendor_reference',
    ]

    @classmethod
    def validate(cls, vendor_info):
        """Validate `vendor_info` to be valid vendor data."""
        assert isinstance(vendor_info, dict)
        assert (vendor_info['type'] in cls.ALLOWED_VENDOR_TYPES)

    @classmethod
    def delete(cls, *args, **kwargs):
        """Disable this endpoint, not valid v2."""
        raise InvalidEndpoint('Not a valid location on this endpoint')

    def remove(self, *args, **kwargs):
        """Disable this operation, not valid on this endpoint."""
        raise InvalidEndpointOperation(
            'Not a valid operation on this endpoint.'
        )

    @classmethod
    def create(cls, data=None, *args, **kwargs):
        """Validate and then create a Vendor entity."""
        cls.validate(data)
        # otherwise endpoint should contain the service path too
        getattr(Entity, 'create').__func__(cls, data=data, *args, **kwargs)



# Copyright (c) PagerDuty.
# See LICENSE for details.
try:
    import ujson as json
except ImportError:
    import json

from .entity import Entity
from .log_entry import LogEntry
from .note import Note
from .alert import Alert
from ..errors import MissingFromEmail


class Incident(Entity):
    """Represents an Incident in PagerDuty's API."""

    STR_OUTPUT_FIELDS = ('id', 'status',)

    logEntryFactory = LogEntry
    noteFactory = Note
    alertFactory = Alert

    def resolve(self, from_email, resolution=None):
        """Resolve an incident using a valid email address."""
        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        endpoint = '/'.join((self.endpoint, self.id,))
        add_headers = {'from': from_email, }
        data = {
            'incident': {
                'type': 'incident',
                'status': 'resolved',
            }
        }

        if resolution is not None:
            data['resolution'] = resolution

        result = self.request('PUT',
                              endpoint=endpoint,
                              add_headers=add_headers,
                              data=data,)
        return result

    def acknowledge(self, from_email):
        """Resolve an incident using a valid email address."""
        endpoint = '/'.join((self.endpoint, self.id,))

        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        add_headers = {'from': from_email, }
        data = {
            'incident': {
                'type': 'incident',
                'status': 'acknowledged',
            }
        }

        result = self.request('PUT',
                              endpoint=endpoint,
                              add_headers=add_headers,
                              data=data,)
        return result

    def log_entries(self, time_zone='UTC', is_overview=False,
                    include=None, fetch_all=True):
        """Query for log entries on an incident instance."""
        endpoint = '/'.join((self.endpoint, self.id, 'log_entries'))

        query_params = {
            'time_zone': time_zone,
            'is_overview': json.dumps(is_overview),
        }

        if include:
            query_params['include'] = include

        result = self.logEntryFactory.find(
            endpoint=endpoint,
            api_key=self.api_key,
            fetch_all=fetch_all,
            **query_params
        )

        return result

    def update(self, *args, **kwargs):
        """Update this incident."""
        raise NotImplemented

    def notes(self):
        """Query for notes attached to this incident."""
        endpoint = '/'.join((self.endpoint, self.id, 'notes'))
        return self.noteFactory.find(
            endpoint=endpoint,
            api_key=self.api_key,
        )

    def create_note(self, from_email, content):
        """Create a note for this incident."""
        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        endpoint = '/'.join((self.endpoint, self.id, 'notes'))
        add_headers = {'from': from_email, }

        return self.noteFactory.create(
            endpoint=endpoint,
            api_key=self.api_key,
            add_headers=add_headers,
            data={'content': content},
        )

    def snooze(self, from_email, duration):
        """Snooze this incident for `duration` seconds."""
        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        endpoint = '/'.join((self.endpoint, self.id, 'snooze'))
        add_headers = {'from': from_email, }

        return self.__class__.create(
            endpoint=endpoint,
            api_key=self.api_key,
            add_headers=add_headers,
            data_key='duration',
            data=3600,
        )

    def merge(self, from_email, source_incidents):
        """Merge other incidents into this incident."""
        if from_email is None or not isinstance(from_email, basestring):
            raise MissingFromEmail(from_email)

        add_headers = {'from': from_email, }
        endpoint = '/'.join((self.endpoint, self.id, 'merge'))
        incident_ids = [entity['id'] if isinstance(entity, Entity) else entity
                        for entity in source_incidents]
        incident_references = [{'type': 'incident_reference', 'id': id_}
                               for id_ in incident_ids]

        return self.__class__.create(
            endpoint=endpoint,
            api_key=self.api_key,
            add_headers=add_headers,
            data_key='source_incidents',
            data=incident_references,
            method='PUT',
        )

    def alerts(self):
        """Query for alerts attached to this incident."""
        endpoint = '/'.join((self.endpoint, self.id, 'alerts'))
        return self.alertFactory.find(
            endpoint=endpoint,
            api_key=self.api_key,
        )
